@@ -15,7 +15,7 @@ import errno
 from . import errors
 from .enums import ICMP_DEFAULT_CODE, IcmpType, IcmpTimeExceededCode, IcmpDestinationUnreachableCode
 
-__version__ = "3.0.2"
+__version__ = "4.0.0"
 DEBUG = False  # DEBUG: Show debug info for developers. (default False)
 EXCEPTIONS = False  # EXCEPTIONS: Raise exception when delay is not available.
 LOGGER = None  # LOGGER: Record logs into console or file.
@@ -161,7 +161,7 @@ def send_one_ping(sock: socket, dest_addr: str, icmp_id: int, seq: int, size: in
     try:
         dest_addr = socket.gethostbyname(dest_addr)  # Domain name will translated into IP address, and IP address leaves unchanged.
     except socket.gaierror as err:
-        raise errors.HostUnknown(dest_addr) from err
+        raise errors.HostUnknown(dest_addr=dest_addr) from err
     _debug("Destination IP address:", dest_addr)
     pseudo_checksum = 0  # Pseudo checksum is used to calculate the real checksum.
     icmp_header = struct.pack(ICMP_HEADER_FORMAT, IcmpType.ECHO_REQUEST, ICMP_DEFAULT_CODE, pseudo_checksum, icmp_id, seq)
@@ -215,7 +215,7 @@ def receive_one_ping(sock: socket, icmp_id: int, seq: int, timeout: int) -> floa
         _debug("Timeout left: {:.2f}s".format(timeout_left))
         selected = select.select([sock, ], [], [], timeout_left)  # Wait until sock is ready to read or time is out.
         if selected[0] == []:  # Timeout
-            raise errors.Timeout(timeout)
+            raise errors.Timeout(timeout=timeout)
         time_recv = time.time()
         _debug("Received time: {} ({}))".format(time.ctime(time_recv), time_recv))
         recv_data, addr = sock.recvfrom(1500)  # Single packet size limit is 65535 bytes, but usually the network packet limit is 1500 bytes.
@@ -223,27 +223,32 @@ def receive_one_ping(sock: socket, icmp_id: int, seq: int, timeout: int) -> floa
             ip_header_raw = recv_data[ip_header_slice]
             ip_header = read_ip_header(ip_header_raw)
             _debug("Received IP header:", ip_header)
+        else:
+            ip_header = None
         icmp_header_raw, icmp_payload_raw = recv_data[icmp_header_slice], recv_data[icmp_header_slice.stop:]
         icmp_header = read_icmp_header(icmp_header_raw)
         _debug("Received ICMP header:", icmp_header)
         _debug("Received ICMP payload:", icmp_payload_raw)
         if not has_ip_header:  # When unprivileged on Linux, ICMP ID is rewrited by kernel.
             icmp_id = sock.getsockname()[1]  # According to https://stackoverflow.com/a/14023878/4528364
-        if icmp_header['id'] and icmp_header['id'] != icmp_id:  # ECHO_REPLY should match the ID field.
-            _debug("ICMP ID dismatch. Packet filtered out.")
-            continue
         if icmp_header['type'] == IcmpType.TIME_EXCEEDED:  # TIME_EXCEEDED has no icmp_id and icmp_seq. Usually they are 0.
-            if icmp_header['code'] == IcmpTimeExceededCode.TTL_EXPIRED:
-                raise errors.TimeToLiveExpired()  # Some router does not report TTL expired and then timeout shows.
+            if icmp_header['code'] == IcmpTimeExceededCode.TTL_EXPIRED:  # Windows raw socket cannot get TTL_EXPIRED. See https://stackoverflow.com/questions/43239862/socket-sock-raw-ipproto-icmp-cant-read-ttl-response.
+                raise errors.TimeToLiveExpired(ip_header=ip_header, icmp_header=icmp_header)  # Some router does not report TTL expired and then timeout shows.
             raise errors.TimeExceeded()
         if icmp_header['type'] == IcmpType.DESTINATION_UNREACHABLE:  # DESTINATION_UNREACHABLE has no icmp_id and icmp_seq. Usually they are 0.
             if icmp_header['code'] == IcmpDestinationUnreachableCode.DESTINATION_HOST_UNREACHABLE:
-                raise errors.DestinationHostUnreachable()
-            raise errors.DestinationUnreachable()
-        if icmp_header['id'] and icmp_header['seq'] == seq:  # ECHO_REPLY should match the SEQ field.
+                raise errors.DestinationHostUnreachable(ip_header=ip_header, icmp_header=icmp_header)
+            raise errors.DestinationUnreachable(ip_header=ip_header, icmp_header=icmp_header)
+        if icmp_header['id']:
             if icmp_header['type'] == IcmpType.ECHO_REQUEST:  # filters out the ECHO_REQUEST itself.
                 _debug("ECHO_REQUEST received. Packet filtered out.")
                 continue
+            if icmp_header['id'] != icmp_id:  # ECHO_REPLY should match the ICMP ID field.
+                _debug("ICMP ID dismatch. Packet filtered out.")
+                continue
+            if icmp_header['seq'] != seq:  # ECHO_REPLY should match the ICMP SEQ field.
+                _debug("IMCP SEQ dismatch. Packet filtered out.")
+                continue
             if icmp_header['type'] == IcmpType.ECHO_REPLY:
                 time_sent = struct.unpack(ICMP_TIME_FORMAT, icmp_payload_raw[0:struct.calcsize(ICMP_TIME_FORMAT)])[0]
                 _debug("Received sent time: {} ({})".format(time.ctime(time_sent), time_sent))
